{ Unit for BTVisualFTP visual FTP client component. }
unit btvftp;

(*

Copyright (C) 2004 Michael Justin

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

*)

interface
{$I IdCompilerDefines.inc}

uses IdFTP, IdFTPCommon, IdFTPList, IdComponent,
{$IFDEF LINUX}
  QComCtrls, QControls, QStdCtrls, QExtCtrls, QImgList, QMenus,
{$ELSE}
  ComCtrls, Controls, StdCtrls, ExtCtrls, Menus,
  Messages, // WM_DROP
{$ENDIF}
  Types, // DWord
  SyncObjs, // TCriticalSection
  Classes;

resourcestring
  rsSkipping = 'Skipping "%s" (file not found)';
  rsUploading = 'Uploading "%s"';
  rsNewFolder = 'Create folder "%s"';
  rsUploadStats = '%0:s%1:s at %2:s kb/sec. (remaining time: %3:s)';
  rsConnectingTo = 'Connecting to %0:s as %1:s';
  rsRemoteDir = 'Remote directory: %0:s';
  rsConnectedTo = 'User %0:s connected to %1:s';
  rsDisconnecting = 'Disconnecting from %0:s';
  rsDisconnected = 'User %0:s disconnected from %1:s';
  rsTransferringData = 'Transferring data for %0:s';
  rsTransferCompleted = 'Transfer completed (%0:s bytes in %1:s files and %2:s folders)';
  rsFileDownloadDone = 'File download complete for "%0:s" (%1:s bytes)';
  rsPutFile = 'Put "%0:s"';
  rsTransferType = 'Transfer type set to %0:s';
  rsNewFolderCaption = 'New folder';
  rsNewFolderName = 'Folder name:';
  rsLoginTo = 'Login to %s';
  rsPasswordFor = 'Password for %s';

  {@exclude confirmation messages }
  rscDeleteFolder = 'Delete folder "%0:s" ?';
  rscDeleteFile = 'Delete File "%0:s" ?';
  rscDeleteSelected = 'Delete %0:s selected items ?';
  rscOverwriteLocal = 'Local file "%0:s" already exists, overwrite?';

  {@exclude error messages }
  rseDeleteFile = 'Could not delete file "%0:s" (Error %1:s)';
  rseDeleteFolder = 'Could not delete folder "%0:s" (Error %1:s)';
  rseNoHost = 'No connection to host.';
  rseCouldNotConnect = 'Could not connect to %0:s';
  rseCouldNotRename = 'Could not rename "%0:s" (Error %1:s)';
  rseCouldNotCWD = 'Could not change to directory %0:s';

  {@exclude listview column captions }
  rslName = 'Name';
  rslSize = 'Size';
  rslDate = 'Date';
  rslType = 'Type';

  {@exclude menu items captions }
  rsiOpen = 'Open';
  rsiGet = 'Get';
  rsiDelete = 'Delete';
  rsiRename = 'Rename';
  rsiNewFolder = 'New folder';
  rsiView = 'View';
  rsiLargeIcons = 'Large icons';
  rsiSmallIcons = 'Small icons';
  rsiList = 'List';
  rsiDetails = 'Details';
  rsiOrder = 'Order';
  rsibyName = 'by Name';
  rsibySize = 'by Size';
  rsibyDate = 'by Date';
  rsibyType = 'by Type';
  rsiTransfertype = 'Transfer type';
  rsiRefresh = 'Refresh';
  rsiBack = 'Back';

  {@exclude other messages }
  rsNotConnected = 'Not connected';
  rsNoDirectory = 'Could not get current directory';
  rsUploadComplete = 'Upload complete';
  rsInProgress = 'File upload in progress!';
  rsTypeFolder = 'Folder';
  rsTypeFile = 'File';
  rsNotAssigned = 'IdFTP must be assigned';
  rsHostMissing = 'IdFTP.Host must be specified';
  rsUserNameMissing = 'IdFTP.Username must be specified';

const
  {@exclude Listview column indices }
  ixSize = 0;
  {@exclude }
  ixDate = 1;
  {@exclude }
  ixType = 2;

type

  TBTVisualFTP = class;

  {@exclude}
  TBTFTPConnection = class(TObject)
    TransferType: TIdFTPTransferType;
    DoneSize, ToBeDoneSize, BytesRead: integer;

  private
    FIdFTP: TIdFTP;
    VisualFTP: TBTVisualFTP;

  public
    procedure SetIdFTP(const Value: TIdFTP);
    function GetIdFTP: TIdFTP;
    function Connect(const RemoteDir: string): boolean;
    function Connected: boolean;
    function PutFile(const Filename: string): boolean;
    function GetFile(const RemoteFile, LocalFile: string;
      const RemoteFilesize: integer): boolean;
    function SetCurrentDirectory(const RemoteDir: string): boolean;
    function GetCurrentDirectory: string;
    function DeleteFile(const Filename: string): boolean;
    function CreateDirectory(const DirName: string): boolean;
    function RemoveDirectory(const Filename: string): boolean;
    function RenameFile(const OldFile, Newfile: string): boolean;
    procedure Disconnect;

    constructor Create(AVisualFTP: TBTVisualFTP);

    property IdFTP: TIdFTP read GetIdFTP write SetIdFTP;
  end;

  // (Storage class for remote file objects.)
  {@exclude}
  TExtraFileData = class(TObject)
  protected
    SizeInBytes: integer;
    IsFolder: boolean;
    IsLink: boolean;
    LinkTarget: string;
{$IFDEF INDY100}
    (**
     * This field is necessary in Indy 10 because on
     * VMS systems, there's a file version feature in the filesystem.
     * The version follows a semicolon after the file extension.  This feature
     * doesn't make sense and the file name version will throw off
     * icon association code.  You use the FileName property for download and upload
     * from the server, not for the icon and associations.
     *)
    LocalFileName: string;
{$ENDIF}
    function IsRoot(const Caption: string): boolean;
    function CanDelete(const Caption: string): boolean;
    function CanRename(const Caption: string): boolean;
  end;

  { Client options. See @link(FTPOptions). }
  TFTPOptions = (
    { Allow to delete files on the remote server. }
    doAllowDelete,
    { Allow to download files from the remote server. }
    doAllowDownload,
    { Allow to create new folders on the remote server. }
    doAllowNewFolder,
    { Allow to rename files on the remote server. }
    doAllowRename,
    { Display upload/download progress. }
    doDisplayProgress,
    { Display hourglass while uploading/downloading files. }
    doShowHourglass,
    { Display ASCII/Binary transfer type. }
    doShowTransferType);

  { Set of client options. See @link(FTPOptions). }
  TdoOptions = set of TFTPOptions;

  TOrderBy = (
    { Order by name. }
    sbName,
    { Order by date. }
    sbDate,
    { Order by size. }
    sbSize,
    { Order by type. }
    sbType);

  {@exclude }
  TDisplayMessageEvent = procedure(Sender: TObject;
    const Messagetext: string;
    const IsError: boolean) of object;

  {@exclude }
  TGetLocalFileEvent = procedure(Sender: TObject;
    const RemoteFilename: string;
    var LocalFileName: string) of object;

  {@exclude }
  TGetConfirmationEvent = procedure(Sender: TObject;
    const Messagetext: string;
    var Confirm: boolean) of object;

  {@exclude }
  TDisplayProgressEvent = procedure(Sender: TObject;
    const DoneSize, ToBeDoneSize, TotalSize: integer;
    const BytesPerSecond: integer;
    const RemoteFilename, LocalFileName: string) of object;

  {@exclude }
  TLocalizeEvent = procedure(Sender: TObject;
    const StrResult: string; var Format: string; const Args: array of const) of
    object;

  { Visual FTP client component. }
  TBTVisualFTP = class(TPanel)
  private
    Connection: TBTFTPConnection;

    TotalSize: Integer;
    StartTime, LastTime: DWord;

    FListView: TListView;
    FInitialRemoteDir: string;
    FUsedDiskSpace: Cardinal;
    FAllowDropFiles: boolean;
    FUploading: boolean;

    FirstItem: Integer;
    FileCount: Integer;
    FolderCount: Integer;

    stateDesc,
      nameDesc,
      ageDesc,
      sizeDesc: boolean;

    OrderBy: TOrderBy;
    ViewMenu, OrderMenu, TransferTypeMenu: TMenuItem;

    UrlPanel: TPanel;
    UrlEdit: TEdit;
    LogSplitter: TSplitter;
    LogWindow: TMemo;

    FShowLogWindow: boolean;
    FShowDirectory: boolean;
    FFTPOptions: TdoOptions;
    FOnDisplayMessage: TDisplayMessageEvent;
    FOnFTPTransferProgress: TDisplayProgressEvent;
    FOnGetConfirmation: TGetConfirmationEvent;
    FOnGetLocalFileName: TGetLocalFileEvent;
    FOnLocalize: TLocalizeEvent;
    FOnChange: TNotifyEvent;
    FOnConnect: TNotifyEvent;
    FOnDisconnect: TNotifyEvent;
    FOnDblClick: TNotifyEvent;

    function GetFileType(const FileName: string; const isFolder: boolean):
      string;
    procedure SetUsedDiskSpace(const Value: Cardinal);
    procedure SetIdFTP(const Value: TIdFTP);
    function GetRemoteDir: string;
    procedure SetRemoteDir(const Value: string);
    procedure Change;
    procedure AddFile(Item: TIdFTPListItem);
    procedure MyOnGetLocalFilename(Sender: TObject;
      const RemoteFilename: string; var LocalFileName: string);
{$IFDEF LINUX}
    procedure MyOnDragOver(Sender, Source: TObject; X, Y: Integer;
      State: TDragState; var Accept: Boolean);
    procedure MyOnDragDrop(Sender, Source: TObject; X, Y: Integer);
{$ENDIF}
    procedure SetAllowDropFiles(const Value: boolean);
{$IFNDEF LINUX}
    procedure WMDropFiles(var msg: TMessage); message WM_DROPFILES;
{$ENDIF}
    procedure SetUploading(const Value: boolean);
    function Upload(Files: string; wait: boolean): boolean;
    procedure SetShowDirectory(const Value: boolean);
    procedure SetListView(const Value: TListView);
{$IFDEF INDY100}
    procedure MyOnWorkBegin(ASender: TObject; AWorkMode: TWorkMode;
      AWorkCountMax: Integer);
    procedure MyOnWork(ASender: TObject; AWorkMode: TWorkMode;
      AWorkCount: Integer);
{$ELSE}
    procedure MyOnWorkBegin(Sender: TObject; AWorkMode: TWorkMode;
      const AWorkCountMax: Integer);
    procedure MyOnWork(Sender: TObject; AWorkMode: TWorkMode;
      const AWorkCount: Integer);
{$ENDIF}

    procedure MyOnWorkEnd(Sender: TObject; AWorkMode: TWorkMode);

{$IFNDEF LINUX}
    procedure ListViewCompare(Sender: TObject; Item1,
      Item2: TListItem; Data: integer; var Compare: integer);
{$ENDIF}
    procedure ListViewColumnClick(Sender: TObject; Column: TListColumn);

{$IFNDEF LINUX}
    procedure ViewLargeIcons(Sender: TObject);
    procedure ViewSmallIcons(Sender: TObject);
{$ENDIF}
    procedure ViewList(Sender: TObject);
    procedure ViewReport(Sender: TObject);

    procedure MyOnURLPanelResize(Sender: TObject);
    procedure MyListviewOnChange(Sender: TObject; Item: TListItem; Change:
      TItemChange);
    procedure MyListViewOnKeyUp(Sender: TObject; var Key: Word; Shift:
      TShiftState);

    procedure OrderByName(Sender: TObject);
    procedure OrderBySize(Sender: TObject);
    procedure OrderByDate(Sender: TObject);
    procedure OrderByType(Sender: TObject);

    function Localize(const Format: string;
      const Args: array of const): string;

    function GetVersion: string;
    procedure SetVersion(const Value: string);

    procedure SetShowLogWindow(const Value: boolean);
    procedure DoDblClick(Sender: TObject);
    procedure SetFTPOptions(Value: TdoOptions);
    procedure SetTransferTypeASCII(Sender: TObject);
    procedure SetTransferTypeBinary(Sender: TObject);
    procedure BackOnClick(Sender: TObject);

{$IFDEF LINUX}
    procedure ListviewEdited(Sender: TObject; Item: TListItem; var S:
      WideString);
{$ELSE}
    procedure ListviewEdited(Sender: TObject; Item: TListItem; var S: string);
{$ENDIF}
    procedure LogSimple(const Messagetext: string);
    function Confirm(const Messagetext: string; const Args: array of const):
      boolean;
    function Rename(const OldName, Newname: string): boolean;
    procedure MyPopupMenuOnPopup(Sender: TObject);
    procedure DeleteOnClick(Sender: TObject);
    procedure MakeDirOnClick(Sender: TObject);
    procedure RenameOnClick(Sender: TObject);
    function NewFolder(const DirName: string): boolean;

    property UsedDiskSpace: Cardinal read FUsedDiskSpace write SetUsedDiskSpace;
    function GetIdFTP: TIdFTP;

  protected
    Lock: TCriticalSection;

    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation);
      override;
    procedure UploadRecursive(const Path: string);
    procedure DisplayProgress(Sender: TObject;
      const DoneSize: integer;
      const RemoteFilename, LocalFileName: string);
    function GetConnection: TBTFTPConnection;
    procedure DoShowErrorMessage(const Messagetext: string; const Args: array of
      const);
    function ShowLastResponseInfo(const wait: boolean): boolean;
    procedure UploadComplete;

  public
    {@exclude}
    constructor Create(AOwner: TComponent); override;
    {@exclude}
    destructor Destroy; override;

    { Connect to the remote server. }
    function Connect: boolean; virtual;
    { Connect from the remote server. }
    procedure Disconnect; virtual;
    { True if a connection has been established. }
    function Connected: boolean; virtual;
    { True if upload is in progress }
    property Uploading: boolean read FUploading write SetUploading;
    { Refresh the file list. }
    procedure RefreshFileList(Sender: TObject);
    { Send text to the log }
    procedure Log(const Messagetext: string; const Args: array of const);

  published
    { Enables drag and drop upload from the Desktop to the server }
    property AllowDropFiles: boolean read FAllowDropFiles write
      SetAllowDropFiles;
    { Sets FTP client rights. }
    property FTPOptions: TdoOptions read FFTPOptions write SetFTPOptions;
    { Link to the Internet Direct (Indy) FTP component. }
    property IdFTP: TIdFTP read GetIdFTP write SetIdFTP;
    { Internal ListView component. }
    property ListView: TListView read FListView write SetListView;
    { Initial remote server directory. }
    property RemoteDir: string read GetRemoteDir write SetRemoteDir;
    { Enables the log window which appears below the ListView. }
    property ShowLogWindow: boolean read FShowLogWindow write SetShowLogWindow;
    { Enables the text field which shows the current remote server directory. }
    property ShowDirectory: boolean read FShowDirectory write SetShowDirectory;
    { The component version. }
    property Version: string read GetVersion write SetVersion stored false;
    { Event which will be called after listview changes. }
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    { Event which will be called after connect. }
    property OnConnect: TNotifyEvent read FOnConnect write FOnConnect;
    { Event which will be called after disconnect. }
    property OnDisconnect: TNotifyEvent read FOnDisconnect write FOnDisconnect;
    { Event which will be called on doubleclick. }
    property OnDblClick: TNotifyEvent read FOnDblClick write FOnDblClick;
    { Event which will be called to get a confirmation. }
    property OnGetConfirmation: TGetConfirmationEvent
      read FOnGetConfirmation write FOnGetConfirmation;
    { Event which will be called to display a message to the user. }
    property OnDisplayMessage: TDisplayMessageEvent
      read FOnDisplayMessage write FOnDisplayMessage;
    { Event which will be called to get a local file name. }
    property OnGetLocalFile: TGetLocalFileEvent
      read FOnGetLocalFileName write FOnGetLocalFileName;
    { Event which will be called to display upload / download progress. }
    property OnDisplayProgress: TDisplayProgressEvent
      read FOnFTPTransferProgress write FOnFTPTransferProgress;
    { Event which will be called to get a translated message string. }
    property OnLocalize: TLocalizeEvent read FOnLocalize write FOnLocalize;

  end;

  // Thread which handles a complete upload sequence
  {@exclude}
  TUploadThread = class(TThread)
  private
    VisualFTP: TBTVisualFTP;
    FileList: TStringList;

  protected
    procedure Execute; override;

  public
    constructor CreateFor(ABTDragDropFTP: TBTVisualFTP; const
      Files: string);

  end;

  { Register the component in the component palette. }
procedure Register;

implementation
uses Math, // Ceil
  IdGlobal, // GPathDelim
{$IFDEF INDY100}
  //needed in Indy 10 because we want all of the list parsers for this
  //Those will link in through a self-registration method.
  IdAllFTPListParsers,
  //required so we can selectively process what the list
  //parser supports.
  IdFTPListParseBase,
{$ENDIF}
  SysUtils,
{$IFDEF LINUX}
  Libc,
  QT, // Key Codes
  QDialogs, // MessageDlg
  QForms; // Application
{$ELSE}
  Dialogs, // MessageDlg
  windows, // TWin32FindData
  wininet, // MAX_PATH_LENGTH
  Registry, // GetFileType
  Forms, // Application
  ShellAPI; // SHGet...
{$ENDIF}

procedure Register;
begin
  RegisterComponents('BetaTools', [TBTVisualFTP]);
end;

{ Util }

{$IFNDEF LINUX}

function GetIconIndex(const AFile: string; Attrs: DWORD): integer;
var
  SFI: TSHFileInfo;
begin
  SHGetFileInfo(PChar(AFile), Attrs, SFI, SizeOf(TSHFileInfo),
    SHGFI_SYSICONINDEX or SHGFI_USEFILEATTRIBUTES);
  Result := SFI.iIcon;
end;
{$ENDIF}

{$IFDEF LINUX}
function GetTickCount: Cardinal;
begin
  result := Ceil(1000*clock/CLOCKS_PER_SEC);
end;
{$ENDIF}

{ TUploadThread }

constructor TUploadThread.CreateFor;
begin
  Create(True);
  Filelist := TStringList.Create;
  Filelist.Text := Files;
  VisualFTP := ABTDragDropFTP;
end;

procedure TUploadThread.Execute;
var
  I: Integer;
begin
  for I := 0 to Filelist.Count - 1 do
  begin
    VisualFTP.Log(Filelist.Text, []);
    VisualFTP.UploadRecursive(Filelist[I]);
  end;
  try
    Filelist.Free;
  finally
    Synchronize(VisualFTP.UploadComplete);
  end;
end;

{ TBTFTPConnection }

constructor TBTFTPConnection.Create;
begin
  inherited Create;
  VisualFTP := AVisualFTP;
end;

function TBTFTPConnection.Connected;
begin
  if Assigned(IdFTP) then
    Result := IdFTP.Connected
  else
    Result := False;
end;

function TBTFTPConnection.Connect;
begin
  if Assigned(IdFTP) then
  try
    IdFTP.Connect(true);

    VisualFTP.Log(IdFTP.Greeting.TextCode
      + ' ' + Trim(IdFTP.Greeting.Text.Text), []);

    VisualFTP.Log(IdFTP.LoginMsg.TextCode
      + ' ' + Trim(IdFTP.LoginMsg.Text.Text), []);

    if RemoteDir = '' then
      Result := True
    else
      Result := SetCurrentDirectory(RemoteDir);
  except
    Result := False;
  end
  else
    Result := False;
end;

function TBTFTPConnection.SetCurrentDirectory;
begin
  if Assigned(IdFTP) then
  try
    IdFTP.ChangeDir(RemoteDir);
    Result := true;
  except
    Result := false;
  end
  else
    Result := False;
end;

function TBTFTPConnection.GetCurrentDirectory: string;
begin
  if Assigned(IdFTP) then
    Result := IdFTP.RetrieveCurrentDir
end;

function TBTFTPConnection.CreateDirectory;
begin
  if Assigned(IdFTP) then
  try
    IdFTP.MakeDir(DirName);
    Result := True;
  except
    result := False;
  end
  else
    Result := False;
end;

function TBTFTPConnection.GetFile;
begin
  if Assigned(IdFTP) then
  try
    IdFTP.Get(RemoteFile, LocalFile);
    Result := True;
  except
    Result := False;
  end
  else
    Result := False;
end;

function TBTFTPConnection.PutFile;
begin
  if Assigned(IdFTP) then
  try
    IdFTP.Put(Filename, ExtractFileName(Filename));
    Result := True;
  except
    Result := False;
  end
  else
    Result := False;
end;

function TBTFTPConnection.DeleteFile;
begin
  if Assigned(IdFTP) then
  try
    IdFTP.Delete(Filename);
    Result := True;
  except
    Result := False;
  end
  else
    Result := False;
end;

function TBTFTPConnection.RenameFile;
begin
  if Assigned(IdFTP) then
  try
    IdFTP.Rename(OldFile, NewFile);
    Result := True;
  except
    Result := False;
  end
  else
    Result := False;
end;

function TBTFTPConnection.RemoveDirectory;
begin
  if Assigned(IdFTP) then
  try
    IdFTP.RemoveDir(Filename);
    Result := True;
  except
    Result := False;
  end
  else
    Result := False;
end;

procedure TBTFTPConnection.Disconnect;
begin
  if Assigned(IdFTP) then
    IdFTP.Disconnect;
end;

procedure TBTFTPConnection.SetIdFTP(const Value: TIdFTP);
begin
  FIdFTP := Value;
end;

function TBTFTPConnection.GetIdFTP: TIdFTP;
begin
  result := FIdFTP;
end;

{ TBTVisualFTP }

{$IFDEF LINUX}

function TBTVisualFTP.GetFileType(const FileName: string; const isFolder:
  boolean): string;
begin
  if isFolder then
    result := Localize(rsTypeFolder, [])
  else
    result := Localize(rsTypeFile, [])
end;
{$ELSE}

function TBTVisualFTP.GetFileType(const FileName: string; const isFolder:
  boolean): string;
var
  TheKey, Ext: string;
begin
  Result := '';
  with TRegistry.Create do
  begin
    RootKey := HKEY_CLASSES_ROOT;
    Ext := ExtractFileExt(Filename);
    if (OpenKey(Ext, False)) or IsFolder then
    begin
      TheKey := ReadString('');
      CloseKey;
      if isFolder then
        TheKey := 'Folder'; // do not localice
      if OpenKey(TheKey, False) then
        Result := ReadString('')
      else
        Result := '';
      CloseKey;
    end
    else
    begin
      system.Delete(Ext, 1, 1);
      Result := UpperCase(Ext) + ' ' + Localize(rsTypeFile, []);
    end;
    Free;
  end;
end;
{$ENDIF}

{$IFNDEF LINUX}
procedure TBTVisualFTP.ViewLargeIcons;
begin
  if not Connected then
    exit;
  ListView.ViewStyle := vsIcon;
  ViewMenu.items[0].Checked := True;
end;

procedure TBTVisualFTP.ViewSmallIcons;
begin
  if not Connected then
    exit;
  ListView.ViewStyle := vsSmallIcon;
  ViewMenu.items[1].Checked := True;
end;
{$ENDIF}

procedure TBTVisualFTP.ViewList;
begin
  if not Connected then
    exit;
  ListView.ViewStyle := vsList;
  ViewMenu.items[2].Checked := True;
end;

procedure TBTVisualFTP.ViewReport;
begin
  if not Connected then
    exit;
  ListView.ViewStyle := vsReport;
  ViewMenu.items[3].Checked := True;
end;

procedure TBTVisualFTP.OrderByName;
begin
  if not Connected then
    exit;
  OrderBy := sbName;
  OrderMenu.items[0].Checked := True;
{$IFDEF LINUX}
  ListView.SortColumn := 0;
{$ELSE}
  ListView.AlphaSort;
{$ENDIF}
end;

procedure TBTVisualFTP.OrderBySize;
begin
  if not Connected then
    exit;
  OrderBy := sbSize;
  OrderMenu.items[1].Checked := True;
{$IFDEF LINUX}
  ListView.SortColumn := 1;
{$ELSE}
  ListView.AlphaSort;
{$ENDIF}
end;

procedure TBTVisualFTP.OrderByDate;
begin
  if not Connected then
    exit;
  OrderBy := sbDate;
  OrderMenu.items[2].Checked := True;
{$IFDEF LINUX}
  ListView.SortColumn := 2;
{$ELSE}
  ListView.AlphaSort;
{$ENDIF}
end;

procedure TBTVisualFTP.OrderByType;
begin
  if not Connected then
    exit;
  OrderBy := sbType;
  OrderMenu.items[3].Checked := True;
{$IFDEF LINUX}
  ListView.SortColumn := 3;
{$ELSE}
  ListView.AlphaSort;
{$ENDIF}
end;

{$IFNDEF LINUX}
procedure TBTVisualFTP.ListViewCompare(Sender: TObject; Item1,
  Item2: TListItem; Data: integer; var Compare: integer);
var
  ci1, ci2: string;
  desc: boolean;
begin
  desc := False;
  Compare := 0;
  try
    case OrderBy of
      sbName:
        begin
          ci1 := UpperCase(Item1.Caption);
          ci2 := UpperCase(Item2.Caption);
          if TExtraFileData(Item1.Data).IsFolder then
            ci1 := ' ' + Item1.Caption;
          if TExtraFileData(Item2.Data).IsFolder then
            ci2 := ' ' + Item2.Caption;
          desc := nameDesc;
          if ci1 < ci2 then
            Compare := -1
          else if ci1 > ci2 then
            Compare := 1;
        end;
      sbDate:
        begin
          if (ixDate < Item1.Subitems.Count) and (ixDate < Item2.Subitems.Count)
            then
          begin
            ci1 := Item1.SubItems[ixDate];
            ci2 := Item2.SubItems[ixDate];
            desc := ageDesc;
            if ci1 = '' then
              ci1 := DateToStr(0);
            if ci2 = '' then
              ci2 := DateToStr(0);
            if StrToDateTime(ci1) < StrToDateTime(ci2) then
              Compare := -1
            else if StrToDateTime(ci1) > StrToDateTime(ci2) then
              Compare := 1;
          end;
        end;
      sbSize:
        begin
          if TExtraFileData(Item1.Data).SizeInBytes <
            TExtraFileData(Item2.Data).SizeInBytes then
            Compare := -1
          else if TExtraFileData(Item1.Data).SizeInBytes >
            TExtraFileData(Item2.Data).SizeInBytes then
            Compare := 1;
          desc := sizeDesc;
        end;
      sbType:
        begin
          ci1 := Item1.SubItems[ixType];
          ci2 := Item2.SubItems[ixType];
          if Item1.SubItems[ixSize] = '' then
            ci1 := ' ' + Item1.Caption;
          if Item2.SubItems[ixSize] = '' then
            ci2 := ' ' + Item2.Caption;
          desc := stateDesc;
          if ci1 < ci2 then
            Compare := -1
          else if ci1 > ci2 then
            Compare := 1;
        end;
    end;

    if desc then
      Compare := Compare * -1;

  except
  end;
end;
{$ENDIF}

function TBTVisualFTP.Localize(const Format: string; const Args: array of
  const):
  string;
var
  f: string;
begin
  result := '';
  f := Format;
  // allow to change the message format
  if Assigned(FOnLocalize) then
    FOnLocalize(self, result, f, Args);
  FmtStr(result, f, Args);
end;

procedure TBTVisualFTP.ListViewColumnClick(Sender: TObject;
  Column: TListColumn);
begin
  if Column.Caption = Localize(rslName, []) then
  begin
    if OrderBy = sbName then
      NameDesc := not NameDesc
    else
      OrderBy := sbName;
  end;

  if Column.Caption = Localize(rslDate, []) then
  begin
    if OrderBy = sbDate then
      ageDesc := not ageDesc
    else
      OrderBy := sbDate;
  end;

  if Column.Caption = Localize(rslSize, []) then
  begin
    if OrderBy = sbSize then
      sizeDesc := not sizeDesc
    else
      OrderBy := sbSize;
  end;

  if Column.Caption = Localize(rslType, []) then
  begin
    if OrderBy = sbType then
      stateDesc := not stateDesc
    else
      OrderBy := sbType;
  end;
{$IFDEF LINUX}
  // ListView.Sort(ListView.SortColumn, sdAscending);
{$ELSE}
  ListView.AlphaSort;
{$ENDIF}
end;

procedure TBTVisualFTP.MyOnURLPanelResize;
begin
  URLEdit.Width := URLPanel.Width;
end;

procedure TBTVisualFTP.MyListviewOnChange;
begin
  if Change = ctState then
  begin
    URLEdit.Text := Connection.GetCurrentDirectory;
  end
end;

procedure TBTVisualFTP.MyListViewOnKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if not Connected then
    exit;
{$IFDEF LINUX}
  case Key of
    Key_Delete: DeleteOnClick(self);
    Key_F2: RenameOnClick(self);
    Key_F5: RefreshFileList(self);
  end;
{$ELSE}
  case Key of
    VK_DELETE: DeleteOnClick(self);
    VK_F2: RenameOnClick(self);
    VK_F5: RefreshFileList(self);
  end;
{$ENDIF}
end;

constructor TBTVisualFTP.Create;
{$IFNDEF LINUX}
var
  SysIL: THandle;
  SysSIL: THandle;
  SFI: TSHFileInfo;
{$ENDIF}
begin
  inherited;

{$IFDEF LINUX}
  Parent := Owner as TWidgetControl;
{$ELSE}
  Parent := Owner as TWinControl;
{$ENDIF}

  Lock := TCriticalSection.Create;

  AllowDropFiles := False;

  Connection := TBTFTPConnection.Create(self);

  URLEdit := TEdit.Create(self);
  with URLEdit do
  begin
    Text := '';
    ReadOnly := true;
  end;

  URLPanel := TPanel.Create(self);
  with URLPanel do
  begin
    Parent := self;
    Align := alTop;
    Caption := '';
    BevelOuter := bvNone;
    BorderWidth := 3;
    Height := (2 * BorderWidth) + URLEdit.Height;
    OnResize := MyOnURLPanelResize;
  end;
  URLEdit.Parent := URLPanel;

  ListView := TListView.Create(self);
  with ListView do
  begin
    Name := 'InternalListView'; // do not localize
    Parent := self;
    ReadOnly := False;
    Align := alClient;
    Viewstyle := vsReport;
    MultiSelect := True;

    OnMouseMove := self.OnMouseMove;
    OnMouseDown := self.OnMouseDown;
    OnMouseUp := self.OnMouseUp;

    OnChange := MyListviewOnChange;
    OnKeyUp := MyListViewOnKeyUp;

    with Columns.Add do
    begin
      Width := 150;
    end;
    with Columns.Add do
    begin
      Width := Canvas.TextWidth('000000000'); // do not localize
      Alignment := taRightJustify;
    end;
    with Columns.Add do
    begin
      Width := Canvas.TextWidth(DateTimeToStr(Now)) + 10;
    end;
    with Columns.Add do
    begin
      Width := 200;
    end;

{$IFDEF LINUX}
    Images := TImageList.Create(self);
    with Images do
    begin
      Name := 'InternalImages'; // do not localize
      Width := 16;
      Height := 16;
    end;
{$ELSE}
    LargeImages := TImageList.Create(self);
    with LargeImages do
    begin
      Name := 'InternalLargeImages'; // do not localize
      Width := 32;
      Height := 32;
      SysIL := ShellAPI.SHGetFileInfo('', 0, SFI, SizeOf(SFI),
        ShellAPI.SHGFI_SYSICONINDEX or ShellAPI.SHGFI_LARGEICON);
      if SysIL <> 0 then
      begin
        LargeImages.Handle := SysIL;
        ShareImages := True;
      end;
    end;

    SmallImages := TImageList.Create(self);
    with SmallImages do
    begin
      Name := 'InternalSmallImages'; // do not localize
      Width := 16;
      Height := 16;
      SysSIL := ShellAPI.SHGetFileInfo('', 0, SFI, SizeOf(SFI),
        ShellAPI.SHGFI_SYSICONINDEX or ShellAPI.SHGFI_SMALLICON);
      if SysSIL <> 0 then
      begin
        SmallImages.Handle := SysSIL;
        ShareImages := True;
      end;
    end;
{$ENDIF}

  end; // Listview

  LogSplitter := TSplitter.Create(self);
  LogSplitter.Parent := self;
  LogSplitter.Align := alBottom;
  LogSplitter.Height := 3;
  LogSplitter.Beveled := False;
  LogSplitter.Cursor := crVSplit;
  LogSplitter.MinSize := 48;

  LogWindow := TMemo.Create(self);
  LogWindow.Parent := self;
  LogWindow.Align := alBottom;
  LogWindow.Height := 64;
  LogWindow.ScrollBars := ssVertical;

  ShowLogWindow := True;
  ShowDirectory := True;

  Height := 250;
  Width := 450;

end;

function TExtraFileData.IsRoot;
begin
  Result := IsFolder and (Caption = '..');
end;

function TExtraFileData.CanDelete;
begin
  Result := not IsRoot(Caption);
end;

function TExtraFileData.CanRename;
begin
  Result := not IsRoot(Caption);
end;

procedure TBTVisualFTP.SetTransferTypeASCII;
begin
  if Assigned(IdFTP) then
    IdFTP.TransferType := ftASCII;
  Log(rsTransferType, ['ASCII']); // do not localize
  TransferTypeMenu.Items[0].Checked := true;
end;

procedure TBTVisualFTP.SetTransferTypeBinary;
begin
  if Assigned(IdFTP) then
    IdFTP.TransferType := ftBinary;
  Log(rsTransferType, ['binary']); // do not localize
  TransferTypeMenu.Items[1].Checked := true;
end;

procedure TBTVisualFTP.MyOnGetLocalFilename(Sender: TObject;
  const RemoteFilename: string;
  var LocalFileName: string);
begin
  {#todo handle cancel }
  with TSaveDialog.Create(self) do
  try
    FileName := RemoteFilename;
    if Execute then
      LocalFileName := FileName;
  finally
    Free;
  end;
end;

procedure TBTVisualFTP.Loaded;
  procedure AddSeparator;
  var
    MenuItem: TMenuItem;
  begin
    MenuItem := TMenuItem.Create(self);
    PopupMenu.items.Add(MenuItem);
    PopupMenu.items[PopupMenu.items.Count - 1].Caption := '-';
  end;

  procedure AddItem(const aCaption: string; aOnClick: TNotifyEvent);
  begin
    PopupMenu.items.Add(TMenuItem.Create(self));
    with PopupMenu.items[PopupMenu.items.Count - 1] do
    begin
      Caption := aCaption;
      OnClick := aOnClick;
    end;
  end;

begin
  inherited;

  if not (csDesigning in ComponentState) then
  begin
    if not Assigned(FOnGetLocalFilename) then
      FOnGetLocalFilename := MyOnGetLocalFilename;

    URLPanel.Visible := FShowDirectory;

    // localize listview columns
    ListView.Columns[0].Caption := Localize(rslName, []);
    ListView.Columns[1].Caption := Localize(rslSize, []);
    ListView.Columns[2].Caption := Localize(rslDate, []);
    ListView.Columns[3].Caption := Localize(rslType, []);

    // build popup menu
    if not Assigned(PopupMenu) then
      PopupMenu := TPopupMenu.Create(self);

    FirstItem := PopupMenu.items.Count;
    PopupMenu.OnPopup := MyPopupMenuOnPopup;

    AddItem('assigned at runtime', DoDblClick); // do not localize
{$IFNDEF LINUX}
    PopupMenu.items[FirstItem].Default := True;
{$ENDIF}
    AddSeparator;
    AddItem(Localize(rsiDelete, []), DeleteOnClick);
    AddItem(Localize(rsiRename, []), RenameOnClick);
    AddItem(Localize(rsiNewFolder, []), MakeDirOnClick);
    ViewMenu := TMenuItem.Create(self);
    PopupMenu.items.Add(ViewMenu);
    with PopupMenu.items[PopupMenu.items.Count - 1] do
    begin
      Caption := Localize(rsiView, []);
      Add(TMenuItem.Create(self));
      items[0].Caption := Localize(rsiLargeIcons, []);
      items[0].RadioItem := True;
{$IFDEF LINUX}
      items[0].Visible := False;
{$ELSE}
      items[0].OnClick := ViewLargeIcons;
{$ENDIF}
      Add(TMenuItem.Create(self));
      items[1].Caption := Localize(rsiSmallIcons, []);
      items[1].RadioItem := True;
{$IFDEF LINUX}
      items[1].Visible := False;
{$ELSE}
      items[1].OnClick := ViewSmallIcons;
{$ENDIF}
      Add(TMenuItem.Create(self));
      items[2].Caption := Localize(rsiList, []);
      items[2].RadioItem := True;
      items[2].OnClick := ViewList;
      Add(TMenuItem.Create(self));
      items[3].Caption := Localize(rsiDetails, []);
      items[3].RadioItem := True;
      items[3].OnClick := ViewReport;
      items[3].Checked := True;
    end;

    AddSeparator;

    OrderMenu := TMenuItem.Create(self);
    PopupMenu.items.Add(OrderMenu);
    with PopupMenu.items[PopupMenu.items.Count - 1] do
    begin
      Caption := Localize(rsiOrder, []);
      Add(TMenuItem.Create(self));
      items[0].Caption := Localize(rsibyName, []);
      items[0].RadioItem := True;
      items[0].Checked := True;
      items[0].OnClick := OrderByName;
      Add(TMenuItem.Create(self));
      items[1].Caption := Localize(rsibySize, []);
      items[1].RadioItem := True;
      items[1].OnClick := OrderBySize;
      Add(TMenuItem.Create(self));
      items[2].Caption := Localize(rsibyDate, []);
      items[2].RadioItem := True;
      items[2].OnClick := OrderByDate;
      Add(TMenuItem.Create(self));
      items[3].Caption := Localize(rsibyType, []);
      items[3].RadioItem := True;
      items[3].OnClick := OrderByType;
    end;

    AddSeparator;

    // Transfer Type
    TransferTypeMenu := TMenuItem.Create(self);
    PopupMenu.items.Add(TransferTypeMenu);
    with PopupMenu.items[PopupMenu.items.Count - 1] do
    begin
      Visible := doShowTransferType in FTPOptions;
      Caption := Localize(rsiTransfertype, []);
      Add(TMenuItem.Create(self));
      items[0].Caption := 'ASCII'; // do not localize
      items[0].RadioItem := True;
      items[0].OnClick := SetTransferTypeASCII;
      Add(TMenuItem.Create(self));
      items[1].Caption := 'Binary'; // do not localize
      items[1].RadioItem := True;
      items[1].Checked := True;
      items[1].OnClick := SetTransferTypeBinary;
    end;

    AddItem(Localize(rsiRefresh, []), RefreshFileList);
    AddSeparator;
    AddItem(Localize(rsiBack, []), BackOnClick);

    ListView.OnColumnClick := self.ListViewColumnClick;
{$IFDEF LINUX}
    ListView.OnDragOver := MyOnDragOver;
    ListView.OnDragDrop := MyOnDragDrop;
{$ELSE}
    ListView.OnCompare := self.ListViewCompare;
{$ENDIF}
    ListView.OnEdited := self.ListviewEdited;
    ListView.OnClick := self.OnClick;
    ListView.OnDblClick := self.DoDblClick;

    if Assigned(IdFTP) then
    begin
      IdFTP.OnWorkBegin := MyOnWorkBegin;
      IdFTP.OnWork := MyOnWork;
      IdFTP.OnWorkEnd := MyOnWorkEnd;
    end;

    // LogSimple(rsNotConnected);

  end;

end;

procedure TBTVisualFTP.SetShowLogWindow(const Value: boolean);
begin
  if Value <> FShowLogWindow then
  begin
    FShowLogWindow := Value;
    LogWindow.Visible := Value;
  end;
end;

procedure TBTVisualFTP.DoDblClick;
var
  RemoteFile, LocalFileName: string;
begin
  if (ListView.Selected <> nil) then
  begin
    RemoteFile := ListView.Selected.Caption;
    if not TExtraFileData(ListView.Selected.Data).IsFolder then
    begin
      LocalFileName := '';
{$IFDEF INDY100}
      //Use the TIdFTPListItem.LocalFileName property for a filename
      //that makes sense on the local system.
      //The reasons are these:
      //1) Unix users prefer lowercase names while systems such as IBM VM/CMS,
      //   VMS, AS/400, and several others are case insensitive.
      //2) VMS has a feature in it's filesystem where more than one version
      //   of a file can be stored.  Thus, if you edit BTVFTP.PAS;1 and save it,
      //   it would be BTVFTP.PAS;2, save it again and it's BTVFTP.PAS;3.
      //   This throws off file association code in the operating system and looks
      //   odd to a user.
      FOnGetLocalFileName(self,
        TExtraFileData(ListView.Selected.Data).LocalFileName, LocalFileName);
{$ELSE}
      FOnGetLocalFileName(self, RemoteFile, LocalFileName);
{$ENDIF}
      if LocalFileName <> '' then
      begin
        if FileExists(LocalFileName) and not Confirm(rscOverwriteLocal,
          [LocalFileName]) then
          exit;
        if Connection.GetFile(RemoteFile, LocalFileName,
          TExtraFileData(ListView.Selected.Data).SizeInBytes) then
          Log(rsFileDownloadDone,
            [RemoteFile,
            IntToStr(TExtraFileData(ListView.Selected.Data).SizeInBytes)])
        else
          ShowLastResponseInfo(False);
      end
    end
    else
    begin
      if Connection.SetCurrentDirectory(RemoteFile) then
        RefreshFileList(self)
      else if not ShowLastResponseInfo(True) then
        DoShowErrorMessage(rseCouldNotCWD, [RemoteFile]);
    end;
  end;
  if Assigned(FOnDblClick) then
    FOnDblClick(Sender)
end;

procedure TBTVisualFTP.SetFTPOptions;
begin
  FFTPOptions := Value;
  // Transfer Type menu visible?
  if Assigned(TransferTypeMenu) then
    TransferTypeMenu.Visible := doShowTransferType in FFTPOptions;
end;

procedure TBTVisualFTP.Log(const Messagetext: string;
  const Args: array of const);
var
  S: string;
begin
  if csDestroying in ComponentState then
    exit;
  Lock.Acquire;
  try
    S := Localize(MessageText, Args);
    if Assigned(FOnDisplayMessage) then
      FOnDisplayMessage(Self, S, False);
    if Assigned(LogWindow) then
      LogWindow.Lines.Add(S);
  finally
    Lock.Release;
  end;
end;

procedure TBTVisualFTP.AddFile(Item: TIdFTPListItem);
var
  Extra: TExtraFileData;
  Filesize: extended;
  procedure FillExtraData;
  begin
    Extra := TExtraFileData.Create;
    Extra.isLink := Item.ItemType = ditSymbolicLink;
{$IFDEF INDY100}
    Extra.LocalFileName := Item.LocalFileName;
{$ENDIF}
    if extra.islink then
    begin
      Extra.LinkTarget := Item.LinkedItemName;
    end;
    Extra.isFolder := (Item.ItemType = ditDirectory) or Extra.IsLink;
  end;
begin
  Application.ProcessMessages;
  if Item.FileName = '.' then
    exit;
  if Item.FileName = '..' then
    exit;

  FillExtraData;
  with Listview.Items.Add do
  begin
    Data := Extra;
    if Extra.isFolder then
    begin
      Inc(FolderCount);
      Caption := PChar(Item.FileName);
{$IFNDEF LINUX}
      ImageIndex := GetIconIndex('.', FILE_ATTRIBUTE_NORMAL or
        FILE_ATTRIBUTE_DIRECTORY);
{$ENDIF}
      subitems.add('');
      Extra.SizeInBytes := -1;
    end
    else
    begin
      Inc(FileCount);
      Caption := Item.FileName;
{$IFNDEF LINUX}
{$IFDEF INDY100}
      ImageIndex := GetIconIndex(Extra.LocalFileName, FILE_ATTRIBUTE_NORMAL);
{$ELSE}
      ImageIndex := GetIconIndex(Caption, FILE_ATTRIBUTE_NORMAL);
{$ENDIF}
{$ENDIF}
{$IFDEF INDY100}
      //In Indy 10, you can't assume file size will be available.  Some parser
      //classes will not return file sizes because the server will not give them
      //and it may be impossible to estimate the file size (or at least, I don't know
      //how to do so.  JPM
      //This is the case on MVS, VSE, and some queues.
      if flcSize in IdFTP.ListParserCapabilities then
      begin
{$ENDIF}
        Filesize := Item.Size / 1024 + 0.0;
        Extra.SizeInBytes := Item.Size;
        if Filesize < 1 then
        begin
          Filesize := 1;
        end;
        subitems.add(Format('%.0n', [Filesize]) + ' KB'); // do not localize
        Inc(FUsedDiskSpace, Item.Size);
{$IFDEF INDY100}
      end
      else
      begin
        //leave the column empty so the user doesn't get an impression
        //of a zero-length file just because we can't get the size.
        subitems.add('');
      end;
{$ENDIF}
    end;
{$IFDEF INDY100}
    //In Indy 10, some parsers can NOT return a date because the list format does not
    //contain a date.  That's the case with HP MPE/iX.  Sometimes, a list parser
    //can not return a date for a particular entry (such as a MVS dataset) while
    //the same parser can return a date for other entries.  We don't want a date
    //to appear as 1899 if we can't get the date.
    if (flcModifiedDate in IdFTP.ListParserCapabilities) and (Item.ModifiedDate
      > 0) then
    begin
      SubItems.Add(DateTimeToStr(Item.ModifiedDate));
    end
    else
    begin
      SubItems.Add('');
    end;
{$ELSE}
    SubItems.Add(DateTimeToStr(Item.ModifiedDate));
{$ENDIF}

{$IFDEF INDY100}
    SubItems.Add(GetFileType(Item.LocalFileName, Extra.isFolder));
{$ELSE}
    SubItems.Add(GetFileType(Item.FileName, Extra.isFolder));
{$ENDIF}

  end;
end;

procedure TBTVisualFTP.RefreshFileList;
var
  I: Integer;
begin
  if not Connected then
    Exit;

  ListView.Items.Clear;
  FUsedDiskSpace := 0;
  FileCount := 0;
  FolderCount := 0;
  Log(rsTransferringData, [Connection.GetCurrentDirectory]);
{$IFNDEF LINUX}
  LockWindowUpdate(ListView.handle);
{$ENDIF}
  ListView.Items.Clear;
  if Connection.GetCurrentDirectory <> '/' then
    with ListView.Items.add do
    begin
      Caption := '..';
{$IFNDEF LINUX}
      ImageIndex := GetIconIndex('.', FILE_ATTRIBUTE_NORMAL or
        FILE_ATTRIBUTE_DIRECTORY);
{$ENDIF}
      SubItems.Add(''); // Size
      SubItems.Add(''); // Date
      SubItems.Add(''); // Type
      Data := TExtraFileData.Create;
      TExtraFileData(Data).IsFolder := True;
    end;

  IdFTP.List(nil);

  for I := 0 to IdFTP.DirectoryListing.Count - 1 do
    AddFile(IdFTP.DirectoryListing.Items[I]);

  Log(rsTransferCompleted, [FormatFloat('#,##0', UsedDiskSpace),
    IntToStr(FileCount), IntToStr(FolderCount)]);
{$IFNDEF LINUX}
  ListView.SortType := stBoth;
  LockWindowUpdate(0);
{$ENDIF}
  Change;
end;

procedure TBTVisualFTP.BackOnClick(Sender: TObject);
begin
  Connection.SetCurrentDirectory('..');
  RefreshFileList(self);
end;

procedure TBTVisualFTP.Change;
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

procedure TBTVisualFTP.ListviewEdited;
begin
  if (not (doAllowRename in FFTPOptions)) or (ListView.Selected.Caption = s)
    then
    Exit;
  Rename(ListView.Selected.Caption, s);
end;

procedure TBTVisualFTP.LogSimple;
begin
  Log(MessageText, []);
end;

function TBTVisualFTP.Confirm;
var
  s: string;
begin
  s := Localize(MessageText, Args);
  if Assigned(FOnGetConfirmation) then
    FOnGetConfirmation(self, s, Result)
  else
    Result := mrYes = MessageDlg(s, mtConfirmation, [mbYes, mbNo], 0);
end;

function TBTVisualFTP.Rename;
begin
  Result := False;
  if Connection.RenameFile(OldName, NewName) then
    with ListView.Selected do
    begin
      Result := True;
      Caption := NewName;
{$IFNDEF LINUX}
      ImageIndex := GetIconIndex(NewName, FILE_ATTRIBUTE_NORMAL);
      SubItems[ixType] := GetFileType(NewName, False);
{$ENDIF}
    end
  else if not ShowLastResponseInfo(True) then
    {#todo1 GetLastError also available in Linux? }
    DoShowErrorMessage(rseCouldNotRename, [OldName, IntToStr(GetLastError)]);
end;

procedure TBTVisualFTP.MyPopupMenuOnPopup;
begin
  if (ListView.Selected <> nil) and
    TExtraFileData(ListView.Selected.Data).IsFolder then
    self.PopupMenu.items[FirstItem].Caption := Localize(rsiOpen, [])
  else
    self.PopupMenu.items[FirstItem].Caption := Localize(rsiGet, []);

  // Don't "get" 'nil' or directories
  self.PopupMenu.items[FirstItem].Enabled := (doAllowDownload in FFTPOptions)
    and Connected and (ListView.Selected <> nil) and (ListView.SelCount = 1);

  // Don't delete 'nil'
  self.PopupMenu.items[FirstItem + 2].Enabled :=
    (doAllowDelete in FFTPOptions) and Connected and (ListView.Selected <> nil)
    and
    TExtraFileData(ListView.Selected.Data).CanDelete(ListView.Selected.Caption);

  // don't rename 'nil'
  self.PopupMenu.items[FirstItem + 3].Enabled :=
    (doAllowRename in FFTPOptions) and Connected and (ListView.Selected <> nil)
    and
    (ListView.SelCount = 1) and TExtraFileData(ListView.Selected.Data).CanRename
    (ListView.Selected.Caption);

  // New folder
  PopupMenu.items[FirstItem + 4].Enabled := (doAllowNewFolder in FFTPOptions)
    and
    Connected;

  // View
  PopupMenu.items[FirstItem + 5].Enabled := Connected;

  // Order
  PopupMenu.items[FirstItem + 7].Enabled := Connected;

  // Refresh
  PopupMenu.items[FirstItem + 10].Enabled := Connected;

  // Back
  PopupMenu.items[FirstItem + 12].Enabled := Connected
    and (Connection.GetCurrentDirectory <> '/');

end;

procedure TBTVisualFTP.MakeDirOnClick;
var
  DirName: string;
begin
  InputQuery(Localize(rsNewFolderCaption, []), Localize(rsNewFolderName, []),
    DirName);
  NewFolder(DirName);
end;

procedure TBTVisualFTP.DeleteOnClick;
var
  i: integer;
  FileName: string;
begin
  if ListView.SelCount = 0 then
    Exit;
  if not (doAllowDelete in FTPOptions) then
    Exit;

  FileName := ListView.Selected.Caption;

  if (ListView.SelCount=1) then
  begin
    if TExtraFileData(ListView.Selected.Data).IsFolder then
    begin
      if not Confirm(rscDeleteFolder, [FileName]) then
        Exit;
    end
    else
    begin
      if not Confirm(rscDeleteFile, [FileName]) then
        Exit;
    end;
  end
  else
  begin
    if not Confirm(rscDeleteSelected, [IntToStr(ListView.SelCount)]) then
      exit;
  end;

  for i := ListView.Selected.Index to ListView.Items.Count - 1 do
    if ListView.Items[i].Selected then
    begin
      FileName := ListView.Items[i].Caption;
      if TExtraFileData(ListView.Items[i].Data).IsFolder then
      begin
        if not Connection.RemoveDirectory(FileName) then
          ShowLastResponseInfo(True);
      end
      else
      begin
        Connection.DeleteFile(FileName);
      end;
    end;

  RefreshFileList(self);
end;

procedure TBTVisualFTP.RenameOnClick;
begin
  if (ListView.Selected <> nil) then
{$IFDEF LINUX}
    ListView.Selected.EditText;
{$ELSE}
    ListView.Selected.EditCaption;
{$ENDIF}
end;

function TBTVisualFTP.NewFolder;
begin
  Result := False;
  if not Connected then
    DoShowErrorMessage(rseNoHost, []);
  if (not Connected) or (Trim(DirName) = '') then
    exit;
  if Connection.CreateDirectory(DirName) then
  begin
    RefreshFileList(self);
    Result := True;
  end
  else
    ShowLastResponseInfo(True);
end;

procedure TBTVisualFTP.DoShowErrorMessage;
var
  s: string;
begin
  s := Localize(MessageText, Args);
  if Assigned(FOnDisplayMessage) then
    FOnDisplayMessage(self, s, True)
  else
    MessageDlg(s, mtError, [mbOK], 0);
end;

{#todo1 return code ? }

function TBTVisualFTP.ShowLastResponseInfo;
begin
  if Assigned(IdFTP) then
    if IdFTP.LastCmdResult.Text.Text <> '' then
    begin
      DoShowErrorMessage(IdFTP.LastCmdResult.Text.Text, []);
      if wait then
        Sleep(200)
    end;
  Result := True;
end;

procedure TBTVisualFTP.SetUsedDiskSpace(const Value: Cardinal);
begin
  FUsedDiskSpace := Value;
end;

procedure TBTVisualFTP.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, OPeration);
  if (Operation = opRemove) then
  begin
    if (AComponent = Connection.IdFTP) then
      Connection.IdFTP := nil;
  end;
end;

procedure TBTVisualFTP.DisplayProgress;
const
  SECONDSPERDAY = 24*60*60;
var
  TobeDoneSize: Integer;
  BytesPerSecond: Integer;
  ElapsedInterval: DWORD;
  ElapsedTime: DWORD;
  DonePercent: Integer;
  KBPerSecond: Currency;
  RemainingTime: TDateTime;
begin
  TobeDoneSize := TotalSize-DoneSize;
  ElapsedInterval := GetTickCount-LastTime;
  if (ElapsedInterval>100) or (ToBeDoneSize=0) then
  begin
    Application.ProcessMessages;
    LastTime := GetTickCount;
    ElapsedTime := GetTickCount-StartTime;

    if ElapsedTime>0 then
      BytesPerSecond := 1000*Ceil(DoneSize/ElapsedTime)
    else
      BytesPerSecond := 1;

    if TotalSize>0 then
      DonePercent := Ceil(100*DoneSize/TotalSize)
    else
      DonePercent := 100;

    if DonePercent=100 then
      KBPerSecond := DoneSize/1024
    else
      KBPerSecond := BytesPerSecond/1024;

    RemainingTime := Ceil(ToBeDoneSize/BytesPerSecond)/SECONDSPERDAY;

    if Assigned(OnDisplayProgress) then
      OnDisplayProgress(self, DoneSize, ToBeDoneSize,
        TotalSize, BytesPerSecond, RemoteFilename, LocalFileName)
    else
      Log(rsUploadStats, [IntToStr(DonePercent), '%', FormatFloat('0.#',
       KBPerSecond), TimeToStr(RemainingTime)]);

  end;

end;

function TBTVisualFTP.GetConnection: TBTFTPConnection;
begin
  result := Connection;
end;

procedure TBTVisualFTP.SetShowDirectory(const Value: boolean);
begin
  if Value <> FShowDirectory then
  begin
    FShowDirectory := Value;
    URLPanel.Visible := Value;
  end;
end;

procedure TBTVisualFTP.SetListView(const Value: TListView);
begin
  FListView := Value;
end;

{$IFDEF INDY100}
procedure TBTVisualFTP.MyOnWorkBegin(ASender: TObject; AWorkMode:
  TWorkMode; AWorkCountMax: Integer);
{$ELSE}

procedure TBTVisualFTP.MyOnWorkBegin(Sender: TObject; AWorkMode:
  TWorkMode; const AWorkCountMax: Integer);
{$ENDIF}
begin
  Application.ProcessMessages;
  if AWorkMode = wmWrite then
  begin
    TotalSize := AWorkCountMax;
    LastTime := GetTickCount;
    StartTime := LastTime;
    DisplayProgress(self, 0, '', '');
  end
end;

{$IFDEF INDY100}

procedure TBTVisualFTP.MyOnWork(ASender: TObject; AWorkMode: TWorkMode;
  AWorkCount: Integer);
{$ELSE}

procedure TBTVisualFTP.MyOnWork(Sender: TObject; AWorkMode: TWorkMode;
  const AWorkCount: Integer);
{$ENDIF}
begin
  if AWorkMode = wmWrite then
    DisplayProgress(self, AWorkCount, '', '');
end;

procedure TBTVisualFTP.MyOnWorkEnd(Sender: TObject; AWorkMode: TWorkMode);
begin
  Application.ProcessMessages;
end;

function TBTVisualFTP.GetIdFTP: TIdFTP;
begin
  result := Connection.IdFTP;
end;

destructor TBTVisualFTP.Destroy;
begin
  if Connected then
    Disconnect;
  Connection.Free;
  inherited;
end;

procedure TBTVisualFTP.SetIdFTP(const Value: TIdFTP);
begin
  Connection.IdFTP := Value;
  // add self to the IdFTP's free notification list
  if Assigned(Value) then
    Value.FreeNotification(Self);
end;

procedure TBTVisualFTP.Disconnect;
begin
  if Assigned(IdFTP) then
  begin
    Log(rsDisconnecting, [IdFTP.Host]);
    ListView.Items.Clear;
    Connection.Disconnect;
    Log(rsDisconnected, [IdFTP.Username, IdFTP.Host]);
  end;
  if not (csDestroying in ComponentState) then
    if Assigned(FOnDisconnect) then
      FOnDisconnect(Self);
end;

function TBTVisualFTP.Connect: boolean;
var
  aPassword: string;
begin
  Result := False;
  if Connected then
    Disconnect;
  if not Assigned(IdFTP) then
    raise Exception.Create(Localize(rsNotAssigned, []));
  if IdFTP.Host = '' then
    raise Exception.Create(Localize(rsHostMissing, []));
  if IdFTP.Username = '' then
    raise Exception.Create(Localize(rsUserNameMissing, []));
  if IdFTP.Password = '' then
  begin
    if not InputQuery(Localize(rsLoginTo, [IdFTP.Host]),
      Localize(rsPasswordFor, [IdFTP.Username]), aPassword) then
      exit;
    IdFTP.Password := aPassword;
  end;

  Log(rsConnectingTo, [IdFTP.Host, IdFTP.Username]);

  if RemoteDir <> '' then
    Log(rsRemoteDir, [RemoteDir]);

  if Connection.Connect(RemoteDir) then
  begin
    Result := True;
    Log(rsConnectedTo, [IdFTP.Username, IdFTP.Host]);
    if Assigned(FOnConnect) then
      FOnConnect(Self);
    RefreshFileList(self);
  end
  else
  begin
    if not ShowLastResponseInfo(True) then
      DoShowErrorMessage(rseCouldNotConnect, [IdFTP.Host]);
    Disconnect;
  end;

{$IFNDEF LINUX}
  // enable drag and drop
  if Result then
    DragAcceptFiles(Handle, AllowDropFiles);
{$ENDIF}

end;

function TBTVisualFTP.GetRemoteDir: string;
begin
  Result := FInitialRemoteDir;
end;

procedure TBTVisualFTP.SetRemoteDir(const Value: string);
begin
  FInitialRemoteDir := Value;
end;

function TBTVisualFTP.Connected: boolean;
begin
  result := Connection.Connected;
end;

procedure TBTVisualFTP.SetUploading(const Value: boolean);
begin
  if Value <> FUploading then
  begin
    FUploading := Value;
{$IFNDEF LINUX}
    if not (csDesigning in ComponentState) then
    begin
      if FUploading then
      begin
        // disable drag and drop while uploading
        DragAcceptFiles(Handle, False);
      end
      else
      begin
        // enable drag and drop depending on flag
        DragAcceptFiles(Handle, AllowDropFiles);
      end
    end
{$ENDIF}
  end
end;

procedure TBTVisualFTP.SetAllowDropFiles(const Value: boolean);
begin
  if Value <> FAllowDropFiles then
  begin
    FAllowDropFiles := Value;
  end;
end;

{$IFNDEF LINUX}

procedure TBTVisualFTP.WMDropFiles;
var
  FileName: PChar;
  Files: string;
  i, Count, size, Drop: integer;
begin
  if not AllowDropFiles then
    exit;

  FileName := '';
  Files := '';
  Application.BringToFront;
  Drop := msg.WParam;
  Count := DragQueryFile(Drop, $FFFFFFFF, '', 0);

  for i := 1 to Count do
  begin
    size := DragQueryFile(Drop, i - 1, nil, 1);
    GetMem(filename, size + 1);
    DragQueryFile(Drop, i - 1, FileName, size + 1);
    Files := Files + FileName + #13#10;
  end;
  DragFinish(Drop);
  FreeMem(filename);

  Upload(Files, False);

  while Uploading do
    Application.ProcessMessages;
end;
{$ENDIF}

function TBTVisualFTP.Upload;
{$IFDEF LINUX}
var
  I: Integer;
{$ENDIF}
begin
  if Uploading then
  begin
    Result := False;
    LogSimple(rsInProgress);
  end
  else
  begin
    if not Connected then
      Connect;
    Uploading := True;
{$IFDEF LINUX}
    {@todo Linux: Thread does not work, upload in main thread instead }
    with TStringlist.Create do
    try
      Text := Files;
      for I := 0 to Count - 1 do
        UploadRecursive(Strings[I]);
    finally
      Free;
    end;
    UploadComplete;
    Result := True;
{$ELSE}
    with TUploadThread.CreateFor(self, Files) do
    try
      FreeOnTerminate := True;
      Resume;
      if Wait then
        WaitFor;
      Result := True; // Thread started
    except;
      Result := False;
    end;
{$ENDIF}
    Uploading := False;
  end
end;

procedure TBTVisualFTP.UploadComplete;
begin
  LogSimple(rsUploadComplete);
  Uploading := False;
  RefreshFileList(self);
end;

{$IFDEF LINUX}

procedure TBTVisualFTP.MyOnDragOver(Sender, Source: TObject; X, Y:
  Integer;
  State: TDragState; var Accept: Boolean);
begin
  Accept := FAllowDropFiles and Connection.Connected;
end;

procedure TBTVisualFTP.MyOnDragDrop(Sender, Source: TObject; X, Y:
  Integer);
var
  MS: TMemoryStream;
  SL: TStrings;
  I: Integer;
begin
  MS := TMemoryStream.Create;
  try
    if SaveDragDataToStream(MS, 'text/uri-list') then
    begin
      MS.Position := 0;
      SL := TStringlist.Create;
      try
        SL.LoadFromStream(MS);
        for I := 0 to SL.Count - 1 do
          if Pos('file:', SL[I]) = 1 then
            SL[I] := Copy(SL[I], 6, Length(SL[I]));
        Upload(SL.Text, false);
      finally
        SL.Free;
      end;
    end;
  finally
    MS.Free;
  end;
end;
{$ENDIF}

procedure TBTVisualFTP.UploadRecursive;
var
  Fullpath: string;

  function MyLogFunction(const Filename: string): boolean;
  begin
    if FileExists(Filename) then
    begin
      Log(rsUploading, [ExtractFileName(Filename)]);
      if not GetConnection.PutFile(Filename) then
      begin
        ShowLastResponseInfo(False);
        Result := False; (* end recursion *)
      end
      else
      begin
        Result := True; (* proceed with recursion *)
      end
    end
    else
    begin
      Log(rsSkipping, [Filename]);
      Result := False; (* end recursion *)
    end;
  end;

  function Recurse(var Path: string; const Mask: string): boolean;
  var
    SRec: TSearchRec;
    Retval: integer;
    Oldlen: integer;
  begin
    Result := True;
    Oldlen := Length(Path);
    (* phase 1, look for normal files *)
    Retval := FindFirst(Path + Mask, faAnyFile, SRec);
    while retval = 0 do
    begin
{$IFDEF LINUX}
      if (SRec.Attr and (faDirectory)) = 0 then
{$ELSE}
      if (SRec.Attr and (faDirectory or faVolumeID)) = 0 then
{$ENDIF}
        (* we found a file, not a directory or volume label,
           log it. Bail out if the log function returns false. *)
        if not MyLogFunction(Path + SRec.Name) then
        begin
          Result := False;
          Break;
        end;
      Retval := FindNext(SRec);
    end;
    Sysutils.FindClose(SRec);

    if not Result then
      Exit;

    (* Phase II, look for subdirectories and recurse thru them *)
    {@todo * under Linux / *.* under Windows ? }
    Retval := FindFirst(Path + '*', faDirectory, SRec);
    while Retval = 0 do
    begin
      if (SRec.Attr and faDirectory) <> 0 then (* we have a directory *)
        if (SRec.Name <> '.') and (SRec.Name <> '..') then
        begin
          Log(rsNewFolder, [SRec.Name]);
          GetConnection.CreateDirectory(SRec.Name);
          GetConnection.SetCurrentDirectory(SRec.Name);
          Path := Path + SRec.Name + IdGlobal.GPathDelim;
          if not Recurse(Path, Mask) then
          begin
            Result := False;
            Break;
          end;
          Delete(Path, Oldlen + 1, 255);
        end;
      Retval := FindNext(SRec);
    end;
    SysUtils.FindClose(SRec);
    GetConnection.SetCurrentDirectory('..');
  end;
begin

  if DirectoryExists(Path) then
  begin
    // it is a folder
    if Path = '' then
    begin
      GetDir(0, Fullpath)
    end
    else
    begin
      Fullpath := Path;
    end;

    GetConnection.CreateDirectory(ExtractFilename(Path));
    GetConnection.SetCurrentDirectory(ExtractFilename(Path));

    if Fullpath[Length(Fullpath)] <> IdGlobal.GPathDelim then
      Fullpath := Fullpath + IdGlobal.GPathDelim;

    Recurse(Fullpath, '*');
  end
  else
  begin
    // it is a file
    MyLogFunction(Path);
  end;

end;

procedure TBTVisualFTP.SetVersion;
begin
  // nothing to do
end;

function TBTVisualFTP.GetVersion: string;
begin
  result := '1.1b4'; // do not localize
end;


end.

